# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=invalid-name
# pylint: disable=missing-docstring

import logging
import os.path as p
import pprint
import tempfile
import time
from contextlib import contextmanager
from multiprocessing import Event
from threading import Thread
from typing import Any, Iterator

import mock

from nose2.tools import such  # type: ignore

from hdl_checker.tests import getTestTempPath, setupTestSuport

from hdl_checker.builder_utils import BuilderName
from hdl_checker.exceptions import UnknownParameterError
from hdl_checker.parsers.config_parser import ConfigParser
from hdl_checker.path import Path
from hdl_checker.types import BuildFlagScope
from hdl_checker.utils import ON_WINDOWS, toBytes

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

SOME_ABS_PATH = "C:\\some\\abs\\path.VHDL" if ON_WINDOWS else "/some/abs/path.VHDL"


@contextmanager
def fileWithContent(content):  # type: (bytes) -> Iterator[str]
    with tempfile.NamedTemporaryFile(delete=False) as fd:
        print("Writing to %s (%s)" % (fd, fd.name))
        fd.write(content)
        fd.flush()
        yield fd.name


such.unittest.TestCase.maxDiff = None

with such.A("config parser object") as it:

    @it.has_setup
    def setup():
        # type: (...) -> Any
        setupTestSuport(TEST_TEMP_PATH)

    #  @it.has_teardown
    #  def teardown():
    #      for temp_path in ('.build', '.hdl_checker'):
    #          temp_path = p.abspath(p.join(CONFIG_PARSER_SUPPORT_PATH,
    #                                       temp_path))
    #          if p.exists(temp_path):
    #              shutil.rmtree(temp_path)

    @it.should(
        "raise UnknownParameterError exception when an unknown " "parameter is found"
    )
    def test_raises_exception():
        # type: (...) -> Any
        with it.assertRaises(UnknownParameterError):
            with fileWithContent(b"foo = bar") as name:
                parser = ConfigParser(Path(name))
                parser.parse()

    with it.having("a regular file"):

        @it.has_setup
        def setup():
            # type: (...) -> Any
            it.path = Path(tempfile.mktemp())

            contents = toBytes(
                """
single_build_flags[vhdl] = -single_build_flag_0
dependencies_build_flags[vhdl] = --vhdl-batch
global_build_flags[vhdl] = -globalvhdl -global-vhdl-flag

single_build_flags[verilog] = -single_build_flag_0
dependencies_build_flags[verilog] = --verilog-batch
global_build_flags[verilog] = -globalverilog -global-verilog-flag

single_build_flags[systemverilog] = -single_build_flag_0
dependencies_build_flags[systemverilog] = --systemverilog-batch
global_build_flags[systemverilog] = -globalsystemverilog -global-systemverilog-flag

builder = msim
target_dir = .build

vhdl work sample_file.vhd -sample_file_flag
vhdl work sample_package.vhdl -sample_package_flag
vhdl work TESTBENCH.VHD -build-in some way

vhdl lib {0}

verilog work foo.v -some-flag some value

systemverilog work bar.sv some sv flag
""".format(
                    SOME_ABS_PATH
                )
            )

            with open(it.path.name, "wb") as fd:
                fd.write(contents)
                fd.flush()

        @it.should("find the correct info")
        def test_parsing_regular_file():
            # type: (...) -> Any
            parser = ConfigParser(it.path)
            config = parser.parse()

            _logger.info("Parsed config:\n%s", pprint.pformat(config))
            sources = config.pop("sources")

            it.assertDictEqual(
                config,
                {
                    "builder": BuilderName.msim.name,
                    "vhdl": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalvhdl",
                                "-global-vhdl-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--vhdl-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    "verilog": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalverilog",
                                "-global-verilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: ("--verilog-batch",),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                    "systemverilog": {
                        "flags": {
                            BuildFlagScope.all.value: (
                                "-globalsystemverilog",
                                "-global-systemverilog-flag",
                            ),
                            BuildFlagScope.dependencies.value: (
                                "--systemverilog-batch",
                            ),
                            BuildFlagScope.single.value: ("-single_build_flag_0",),
                        }
                    },
                },
            )

            def _resolve(path):
                return p.join(it.path.dirname, path)

            it.assertCountEqual(
                sources,
                [
                    (
                        _resolve("sample_file.vhd"),
                        {"library": "work", "flags": ("-sample_file_flag",)},
                    ),
                    (
                        _resolve("sample_package.vhdl"),
                        {"library": "work", "flags": ("-sample_package_flag",)},
                    ),
                    (
                        _resolve("TESTBENCH.VHD"),
                        {"library": "work", "flags": ("-build-in", "some", "way")},
                    ),
                    (SOME_ABS_PATH, {"library": "lib", "flags": ()}),
                    (
                        _resolve("foo.v"),
                        {"library": "work", "flags": ("-some-flag", "some", "value")},
                    ),
                    (
                        _resolve("bar.sv"),
                        {"library": "work", "flags": ("some", "sv", "flag")},
                    ),
                ],
            )

        @it.should("only parse when the file actually changes")
        def test_only_parse_when_source_changes():
            # type: (...) -> Any
            parser = ConfigParser(it.path)

            _logger.info("Parsing %s for the 1st time", it.path)
            with mock.patch.object(parser, "_parseLine") as _parseLine:
                parser.parse()
                _parseLine.assert_called()

            time.sleep(0.1)

            _logger.info("Parsing %s for the 2nd time", it.path)
            with mock.patch.object(parser, "_parseLine") as _parseLine:
                parser.parse()
                _parseLine.assert_not_called()

        @it.should("report parsing in progress")
        def test_should_report_parsing_in_progress():
            # type: (...) -> Any
            parser = ConfigParser(it.path)
            it.assertFalse(parser.isParsing(), "Parser should not be busy right now")

            event = Event()

            # Make parsing wait until we set the event to mitigate race
            # conditions
            def _parse(*args, **kwargs):
                it.assertTrue(event.wait(timeout=3), "Timeout waiting for event")

            with mock.patch.object(parser, "_parse", _parse):
                thread = Thread(target=parser.parse)
                thread.start()
                time.sleep(0.1)
                it.assertTrue(
                    parser.isParsing(), "Parser should indicate it's still parsing"
                )
                event.set()


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Top of the hdl_checker.parsers submodule"

import json
import logging
import os
import os.path as p
import subprocess as subp
from glob import iglob as glob
from typing import Any, Dict, Iterable, NamedTuple, Optional, Set, Tuple, Type, Union

import six

from .parsers.verilog_parser import VerilogParser
from .parsers.vhdl_parser import VhdlParser

from hdl_checker import DEFAULT_PROJECT_FILE
from hdl_checker.exceptions import UnknownTypeExtension
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, BuildFlagScope, FileType
from hdl_checker.utils import ON_WINDOWS, isFileReadable, toBytes

_logger = logging.getLogger(__name__)

PARSERS = {
    FileType.vhdl: VhdlParser,
    FileType.verilog: VerilogParser,
    FileType.systemverilog: VerilogParser,
}  # type: Dict[FileType, Type[Union[VhdlParser, VerilogParser]]]


def getSourceParserFromPath(path):  # type: (Path) -> Union[VhdlParser, VerilogParser]
    """
    Returns either a VhdlParser or VerilogParser based on the path's file
    extension
    """
    return PARSERS[FileType.fromPath(path)](path)


def _makeAbsoluteIfNeeded(root, paths):
    # type: (str, Iterable[str]) -> Iterable[str]
    "Makes paths absolute by prepending root if needed"
    for path in paths:
        if p.isabs(path):
            yield path
        else:
            yield p.join(root, path)


def getIncludedConfigs(search_paths, root_dir="."):
    # type: (Iterable[str], str) -> Iterable[Tuple[str, Dict[str, Any]]]
    "Returns configuration contents of included files"
    # Copy the dict to avoid messing up with the caller's env

    # Will search for inclusion clauses recursivelly but we need to keep
    # track of infinite loops
    checked_paths = set()  # type: Set[str]

    paths = set(_makeAbsoluteIfNeeded(root_dir, search_paths))

    while paths:
        path = paths.pop()
        checked_paths.add(path)

        if not p.exists(path):
            _logger.warning("Skipping included path '%s' (no such file)", path)
            continue

        use_json = False

        # Cover cases where 'path' is an existing file
        if not p.isdir(path):
            use_json = True

        # Cover cases where 'path' is a folder that has a file named
        # DEFAULT_PROJECT_FILE inside
        if p.isdir(path) and p.exists(p.join(path, DEFAULT_PROJECT_FILE)):
            path = p.join(path, DEFAULT_PROJECT_FILE)
            _logger.debug("Path is a folder and %s exists, using it", path)
            use_json = True

        if use_json:
            # Load the config from the file
            try:
                config = json.load(open(path, "r"))
            except json.decoder.JSONDecodeError:
                _logger.warning("Failed to decode file %s", path)
                continue

            extracted_paths = set(
                _makeAbsoluteIfNeeded(p.dirname(path), config.pop("include", ()))
            )
            # Add new paths to the set so they get parsed as well
            paths |= extracted_paths - checked_paths

            yield p.dirname(path), config
        else:
            # If the path pointed to a folder, search it for sources
            yield path, {"sources": tuple(findRtlSourcesByPath(Path(path)))}


class JsonSourceEntry(
    NamedTuple(
        "JsonSourceEntry",
        (("path_expr", str), ("library", Optional[str]), ("flags", str)),
    )
):
    """
    Converts different methods of representing a source on the JSON file to a
    saner named tuple
    """

    @classmethod
    def make(cls, iterable):  # pylint: disable=arguments-differ
        # type: (...) -> Any
        """
        Creates a JsonSourceEntry from all supported formats:
            - str
            - [str, {"library": "<library_name>", "flags": BuildFlags}]
            - {"path": str,               <- Mandatory
               "library": <library_name>, <- Optional
               "flags": <BuildFlags>,     <- Optional
              }
        """
        path = iterable
        info = {}  # type: Dict[str, Union[None, str, BuildFlags]]

        # Support both
        if not isinstance(path, six.string_types):
            path = iterable[0]
            info = iterable[1]

        library = info.get("library", None)
        flags = info.get("flags", tuple())

        return super(JsonSourceEntry, cls)._make([path, library, flags])


SourceEntry = NamedTuple(
    "SourceEntry",
    (
        ("path", Path),
        ("library", Optional[str]),
        ("source_specific_flags", BuildFlags),
        ("single_flags", BuildFlags),
        ("dependencies_flags", BuildFlags),
    ),
)


def flattenConfig(root_config, root_path):
    # type: (Dict[str, Any], str) -> Iterable[SourceEntry]
    """
    Expands the given root config and also recursively expands included JSON
    files as well
    """
    # Extract included paths and yield those results first
    include_paths = root_config.pop("include", ())
    for config_path, config in getIncludedConfigs(include_paths, root_path):
        for entry in _expand(config, config_path):
            yield entry

    # Extract sources form the root config last so these are the values that
    # might eventually prevail
    for entry in _expand(root_config, root_path):
        yield entry


def _expand(config, ref_path):
    # type: (Dict[str, Any], str) -> Iterable[SourceEntry]
    """
    Expands the sources defined in the config dict into a list of tuples
    """

    flags = {}

    for filetype in FileType:
        filetype_cfg = config.pop(filetype.value, {}).pop("flags", {})
        flags[filetype] = (
            filetype_cfg.get(BuildFlagScope.single.value, ()),
            filetype_cfg.get(BuildFlagScope.dependencies.value, ()),
            filetype_cfg.get(BuildFlagScope.all.value, ()),
        )

    sources = config.pop("sources", None)

    # If no sources were defined, search ref_path
    if sources is None:
        _logger.debug("No sources found, will search %s", ref_path)
        sources = (x.name for x in findRtlSourcesByPath(Path(ref_path)))

    for entry in sources:
        source = JsonSourceEntry.make(entry)
        path_expr = (
            source.path_expr
            if p.isabs(source.path_expr)
            else p.join(ref_path, source.path_expr)
        )

        for _path in glob(path_expr, recursive=True):
            path = Path(_path, ref_path)

            try:
                filetype = FileType.fromPath(path)
            except UnknownTypeExtension:
                _logger.warning("Won't include non RTL file '%s'", path)
                continue

            single_flags = flags[filetype][0]
            dependencies_flags = flags[filetype][1]
            global_flags = flags[filetype][2]

            yield SourceEntry(
                path,
                source.library,
                tuple(source.flags),
                tuple(global_flags) + tuple(single_flags),
                tuple(global_flags) + tuple(dependencies_flags),
            )


def findRtlSourcesByPath(path):
    # type: (Path) -> Iterable[Path]
    """
    Finds RTL sources (files with extensions within FileType enum) inside
    <path>
    """
    for dirpath, _, filenames in os.walk(path.name):
        for filename in filenames:
            full_path = Path(p.join(dirpath, filename))

            if not p.isfile(full_path.name):
                continue

            try:
                # FileType.fromPath will fail if the file's extension is not
                # valid (one of '.vhd', '.vhdl', '.v', '.vh', '.sv',
                # '.svh')
                FileType.fromPath(full_path)
            except UnknownTypeExtension:
                continue

            if isFileReadable(full_path.name):
                yield full_path


def isGitRepo(path):
    # type: (Path) -> bool
    """
    Checks if path is a git repository by checking if 'git -C path rev-parse
    --show-toplevel' returns an existing path
    """
    cmd = ("git", "-C", path.abspath, "rev-parse", "--show-toplevel")

    try:
        return p.exists(subp.check_output(cmd, stderr=subp.STDOUT).decode().strip())
    except (subp.CalledProcessError, FileNotFoundError):
        return False


def _gitLsFiles(path_to_repo, recurse_submodules=False):
    # type: (Path, bool) -> Iterable[Path]
    "Lists files from a git repository"
    cmd = ["git", "-C", path_to_repo.abspath, "ls-files"]
    if recurse_submodules:
        cmd += ["--recurse-submodules"]

    for line in subp.check_output(cmd, stderr=subp.STDOUT).decode().split("\n"):
        if not line:
            continue
        yield Path(line, path_to_repo.abspath)


def _filterGitIgnoredPathsOnWin(path_to_repo, paths):
    # type: (Path, Iterable[Path]) -> Iterable[Path]
    """
    Filters out paths that are ignored by git; paths outside the repo are kept.
    Uses a multiple calls to 'git check-ignore' and checks if the output is
    empty (ignored paths will be echoed). The command will return non zero exit
    code for paths outside the repo or if <path_to_repo> is not actually a git
    repo, in which cases paths will be included.
    """
    files = _gitLsFiles(path_to_repo, recurse_submodules=True)
    base_cmd = ("git", "-C", path_to_repo.abspath, "check-ignore")

    for path in paths:
        # If the file was found on git ls-files, it is part of the repo and
        # there's no need to check if any git ignore rules match
        if path in files:
            yield path
            continue

        cmd = base_cmd + (str(path),)
        try:
            if not subp.check_output(cmd, stderr=subp.STDOUT):
                yield path
        except subp.CalledProcessError:
            yield path


def _filterGitIgnoredPathsOnUnix(path_to_repo, paths):
    # type: (Path, Iterable[Path]) -> Iterable[Path]
    """
    Filters out paths that are ignored by git; paths outside the repo are kept.
    Uses a 'git check-ignore --stdin' and writes <paths> iteratively to avoid
    piping to the OS all the time
    """
    _logger.debug("Filtering git ignored files from %s", path_to_repo)

    cmd = (
        "git",
        "-C",
        path_to_repo.abspath,
        "check-ignore",
        "--verbose",
        "--non-matching",
        "--stdin",
    )

    proc = None

    for path in paths:
        # Besides the first iteration, the process also needs to be recreated
        # whenever it has died
        if proc is None:
            proc = subp.Popen(cmd, stdin=subp.PIPE, stdout=subp.PIPE, stderr=subp.PIPE)

        proc.stdin.write(toBytes(str(path.abspath) + "\n"))
        # Flush so that data makes to the process
        proc.stdin.flush()

        if proc.stdout.readline().decode().startswith("::"):
            yield path

        # proc will die whenever we write a path that's outside the repo.
        # Because this method aims to filter *out* ignored files and files
        # outside the repo aren't subject to this filter, we'll include them
        if proc.poll() is not None:
            yield path
            # Deallocate the process (hopefully this won't leave a zombie
            # process behind)
            del proc
            proc = None


def filterGitIgnoredPaths(path_to_repo, paths):
    # type: (Path, Iterable[Path]) -> Iterable[Path]
    """
    Filters out paths that are ignored by git; paths outside the repo are kept.
    """
    files = set(_gitLsFiles(path_to_repo, recurse_submodules=True))
    paths = set(paths)

    # Files found on git ls-files are part of the repo, no need to check if any
    # git ignore rules match
    for path in files.intersection(paths):
        yield path

    func = _filterGitIgnoredPathsOnWin if ON_WINDOWS else _filterGitIgnoredPathsOnUnix

    for path in func(path_to_repo, paths - files):
        yield path



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base class that implements the base builder flow"

import logging

from .simple_finder import SimpleFinder

_logger = logging.getLogger(__name__)


def getGeneratorByName(name):
    "Returns the builder class given a string name"
    # Check if the builder selected is implemented and create the
    # builder attribute
    return {"SimpleFinder": SimpleFinder}.get(name, SimpleFinder)


__all__ = ["SimpleFinder"]



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base source file parser"

import abc
import logging
import os.path as p
from typing import Any, Dict, Optional, Set

from .elements.dependency_spec import (
    BaseDependencySpec,
)  # pylint: disable=unused-import
from .elements.design_unit import tAnyDesignUnit  # pylint: disable=unused-import

from hdl_checker.path import Path
from hdl_checker.types import FileType
from hdl_checker.utils import HashableByKey, readFile, removeDuplicates

_logger = logging.getLogger(__name__)


class BaseSourceFile(HashableByKey):  # pylint:disable=too-many-instance-attributes
    """
    Parses and stores information about a source file such as design
    units it depends on and design units it provides
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, filename):
        # type: (Path, ) -> None
        assert isinstance(filename, Path), "Invalid type: {}".format(filename)
        self.filename = filename
        self._cache = {}  # type: Dict[str, Any]
        self._content = None  # type: Optional[str]
        self._mtime = 0  # type: Optional[float]
        self.filetype = FileType.fromPath(self.filename)
        self._dependencies = None  # type: Optional[Set[BaseDependencySpec]]
        self._design_units = None  # type: Optional[Set[tAnyDesignUnit]]
        self._libraries = None

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        state = self.__dict__.copy()
        del state["_content"]
        del state["_design_units"]
        del state["_dependencies"]
        return state

    @classmethod
    def __jsonDecode__(cls, state):
        """
        Returns an object of cls based on a given state
        """
        obj = super(BaseSourceFile, cls).__new__(cls)
        obj.filename = state["filename"]
        obj.filetype = state["filetype"]
        obj._cache = state["_cache"]  # pylint: disable=protected-access
        obj._content = None  # pylint: disable=protected-access
        obj._mtime = state["_mtime"]  # pylint: disable=protected-access
        obj._dependencies = None  # pylint: disable=protected-access
        obj._design_units = None  # pylint: disable=protected-access
        obj._libraries = state["_libraries"]  # pylint: disable=protected-access

        return obj

    def __repr__(self):
        return "{}(filename={}, design_units={}, dependencies={})".format(
            self.__class__.__name__,
            self.filename,
            self._design_units,
            self._dependencies,
        )

    @property
    def __hash_key__(self):
        return (self.filename, self._content)

    def _changed(self):
        # type: (...) -> Any
        """
        Checks if the file changed based on the modification time
        provided by p.getmtime
        """
        if not p.exists(str(self.filename)):
            return False
        return bool(self.getmtime() > self._mtime)  # type: ignore

    def _clearCachesIfChanged(self):
        # type: () -> None
        """
        Clears all the caches if the file has changed to force updating
        every parsed info
        """
        if self._changed():
            self._content = None
            self._dependencies = None
            self._design_units = None
            self._libraries = None
            self._cache = {}

    def getmtime(self):
        # type: () -> Optional[float]
        """
        Gets file modification time as defined in p.getmtime
        """
        if not p.exists(self.filename.name):
            return None
        return self.filename.mtime

    def getSourceContent(self):
        # type: (...) -> Any
        """
        Cached version of the _getSourceContent method
        """
        self._clearCachesIfChanged()

        if self._content is None:
            self._content = self._getSourceContent()
            self._mtime = self.getmtime()

        return self._content

    def _getSourceContent(self):
        # type: () -> str
        """
        Method that can be overriden to change the contents of the file, like
        striping comments off.
        """
        return readFile(str(self.filename))

    def getDesignUnits(self):  # type: () -> Set[tAnyDesignUnit]
        """
        Cached version of the _getDesignUnits method
        """
        if not p.exists(self.filename.name):
            return set()
        self._clearCachesIfChanged()
        if self._design_units is None:
            self._design_units = set(self._getDesignUnits())

        return self._design_units

    def getDependencies(self):
        # type: () -> Set[BaseDependencySpec]
        """
        Cached version of the _getDependencies method
        """
        if not p.exists(self.filename.name):
            return set()

        self._clearCachesIfChanged()
        if self._dependencies is None:
            try:
                self._dependencies = set(self._getDependencies())
            except:  # pragma: no cover
                print("Failed to parse %s" % self.filename)
                _logger.exception("Failed to parse %s", self.filename)
                raise

        return self._dependencies

    def getLibraries(self):
        # type: (...) -> Any
        """
        Cached version of the _getLibraries method
        """
        if not p.exists(self.filename.name):
            return []

        self._clearCachesIfChanged()
        if self._libraries is None:
            self._libraries = removeDuplicates(self._getLibraries())

        return self._libraries

    @abc.abstractmethod
    def _getDesignUnits(self):
        """
        Method that should implement the real parsing of the source file
        to find design units defined. Use the output of the getSourceContent
        method to avoid unnecessary I/O
        """

    def _getLibraries(self):
        """
        Parses the source file to find libraries required by the file
        """
        return ()

    @abc.abstractmethod
    def _getDependencies(self):
        """
        Parses the source and returns a list of dictionaries that
        describe its dependencies
        """



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"VHDL source file parser"

import logging
import re
import string
from typing import Any, Generator, Iterable, List, Tuple, Type

from .elements.dependency_spec import (
    BaseDependencySpec,
    IncludedPath,
    RequiredDesignUnit,
)
from .elements.design_unit import VerilogDesignUnit
from .elements.parsed_element import Location

from hdl_checker.parsers.base_parser import BaseSourceFile
from hdl_checker.parsers.elements.identifier import VerilogIdentifier
from hdl_checker.types import DesignUnitType, FileType
from hdl_checker.utils import readFile

_logger = logging.getLogger(__name__)

_VERILOG_IDENTIFIER = "|".join(
    [r"[a-zA-Z_][a-zA-Z0-9_$]+", r"\\[%s]+(?=\s)" % string.printable.replace(" ", "")]
)

_COMMENT = r"(?:\/\*.*?\*\/|//[^(?:\r\n?|\n)]*)"


# Design unit scanner
_DESIGN_UNITS = re.compile(
    "|".join(
        [
            r"(?<=\bmodule\b)\s*(?P<module_name>%s)" % _VERILOG_IDENTIFIER,
            r"(?<=\bpackage\b)\s*(?P<package_name>%s)" % _VERILOG_IDENTIFIER,
            _COMMENT,
        ]
    ),
    flags=re.DOTALL,
)

_DEPENDENCIES = re.compile(
    "|".join(
        [
            r"(?P<package>\b{0})\s*::\s*(?:{0}|\*)".format(_VERILOG_IDENTIFIER),
            r"(\bvirtual\b)?\s*\bclass\s+(?:static|automatic)?(?P<class>\b{0})".format(
                _VERILOG_IDENTIFIER
            ),
            r"(?<=`include\b)\s*\"(?P<include>.*?)\"",
            _COMMENT,
        ]
    ),
    flags=re.DOTALL,
)


class VerilogParser(BaseSourceFile):
    """
    Parses and stores information about a Verilog or SystemVerilog
    source file
    """

    def _getSourceContent(self):
        # type: (...) -> Any
        # Remove multiline comments
        content = readFile(str(self.filename))
        return content
        #  lines = _COMMENT.sub("", content)
        #  return re.sub(r"\r\n?|\n", " ", lines, flags=re.S)

    def _iterDesignUnitMatches(self):
        # type: (...) -> Any
        """
        Iterates over the matches of _DESIGN_UNITS against
        source's lines
        """
        content = self.getSourceContent()
        lines = content.split("\n")
        for match in _DESIGN_UNITS.finditer(self.getSourceContent()):
            start = match.start()
            start_line = content[:start].count("\n")

            total_chars_to_line_with_match = len("\n".join(lines[:start_line]))
            start_char = match.start() - total_chars_to_line_with_match

            yield match.groupdict(), {Location(start_line, start_char)}

    def _getDependencies(self):  # type: () -> Iterable[BaseDependencySpec]
        text = self.getSourceContent()

        match_groups = [
            ("include", IncludedPath)
        ]  # type: List[Tuple[str, Type[BaseDependencySpec]]]

        # Only SystemVerilog has imports or classes
        if self.filetype is FileType.systemverilog:
            match_groups += [
                ("package", RequiredDesignUnit),
                ("class", RequiredDesignUnit),
            ]

        for match in _DEPENDENCIES.finditer(text):
            for match_group, klass in match_groups:
                name = match.groupdict().get(match_group, None)
                # package 'std' seems to be built-in. Need to have a look a
                # this if include_name is not None and include_name != 'std':
                if match_group == "package" and name == "std":
                    continue

                # package 'std' seems to be built-in. Need to have a look a this
                if name is None:
                    continue

                line_number = text[: match.end()].count("\n")
                column_number = len(text[: match.start()].split("\n")[-1])

                yield klass(
                    owner=self.filename,
                    name=VerilogIdentifier(name),
                    locations=(Location(line_number, column_number),),
                )
                break

    def _getDesignUnits(self):  # type: () -> Generator[VerilogDesignUnit, None, None]
        for match, locations in self._iterDesignUnitMatches():
            if match["module_name"] is not None:
                yield VerilogDesignUnit(
                    owner=self.filename,
                    name=match["module_name"],
                    type_=DesignUnitType.entity,
                    locations=locations,
                )
            if match["package_name"] is not None:
                yield VerilogDesignUnit(
                    owner=self.filename,
                    name=match["package_name"],
                    type_=DesignUnitType.package,
                    locations=locations,
                )
